from __future__ import annotations

import asyncio
import contextlib
import json
import time
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from warnings import warn

from httpx import HTTPStatusError
from starlette.datastructures import Headers
from starlette.requests import Request
from starlette.routing import BaseRoute, Match, Router
from starlette.schemas import EndpointInfo, SchemaGenerator
from starlette.testclient import TestClient
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from apitally.client.asyncio import ApitallyClient
from apitally.client.base import Consumer as ApitallyConsumer
from apitally.common import get_versions


__all__ = ["ApitallyMiddleware", "ApitallyConsumer"]


class ApitallyMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        client_id: str,
        env: str = "dev",
        app_version: Optional[str] = None,
        openapi_url: Optional[str] = "/openapi.json",
        filter_unhandled_paths: bool = True,
        identify_consumer_callback: Optional[Callable[[Request], Union[str, ApitallyConsumer, None]]] = None,
    ) -> None:
        self.app = app
        self.filter_unhandled_paths = filter_unhandled_paths
        self.identify_consumer_callback = identify_consumer_callback
        self.client = ApitallyClient(client_id=client_id, env=env)
        self.client.start_sync_loop()
        self._delayed_set_startup_data_task: Optional[asyncio.Task] = None
        self.delayed_set_startup_data(app_version, openapi_url)
        _register_shutdown_handler(app, self.client.handle_shutdown)

    def delayed_set_startup_data(self, app_version: Optional[str] = None, openapi_url: Optional[str] = None) -> None:
        self._delayed_set_startup_data_task = asyncio.create_task(
            self._delayed_set_startup_data(app_version, openapi_url)
        )

    async def _delayed_set_startup_data(
        self, app_version: Optional[str] = None, openapi_url: Optional[str] = None
    ) -> None:
        await asyncio.sleep(1.0)  # Short delay to allow app routes to be registered first
        data = _get_startup_data(self.app, app_version, openapi_url)
        self.client.set_startup_data(data)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "http" and scope["method"] != "OPTIONS":
            request = Request(scope)
            response_status = 0
            response_time = 0.0
            response_headers = Headers()
            response_body = b""
            response_size = 0
            response_chunked = False
            start_time = time.perf_counter()

            async def send_wrapper(message: Message) -> None:
                nonlocal \
                    response_time, \
                    response_status, \
                    response_headers, \
                    response_body, \
                    response_size, \
                    response_chunked
                if message["type"] == "http.response.start":
                    response_time = time.perf_counter() - start_time
                    response_status = message["status"]
                    response_headers = Headers(scope=message)
                    response_chunked = (
                        response_headers.get("Transfer-Encoding") == "chunked"
                        or "Content-Length" not in response_headers
                    )
                elif message["type"] == "http.response.body":
                    if response_chunked:
                        response_size += len(message.get("body", b""))
                    if response_status == 422:
                        response_body += message.get("body", b"")
                await send(message)

            try:
                await self.app(scope, receive, send_wrapper)
            except BaseException as e:
                self.add_request(
                    request=request,
                    response_status=500,
                    response_time=time.perf_counter() - start_time,
                    response_headers=response_headers,
                    response_body=response_body,
                    response_size=response_size,
                    exception=e,
                )
                raise e from None
            else:
                self.add_request(
                    request=request,
                    response_status=response_status,
                    response_time=response_time,
                    response_headers=response_headers,
                    response_body=response_body,
                    response_size=response_size,
                )
        else:
            await self.app(scope, receive, send)  # pragma: no cover

    def add_request(
        self,
        request: Request,
        response_status: int,
        response_time: float,
        response_headers: Headers,
        response_body: bytes,
        response_size: int = 0,
        exception: Optional[BaseException] = None,
    ) -> None:
        path_template, is_handled_path = self.get_path_template(request)
        if is_handled_path or not self.filter_unhandled_paths:
            consumer = self.get_consumer(request)
            consumer_identifier = consumer.identifier if consumer else None
            self.client.consumer_registry.add_or_update_consumer(consumer)
            self.client.request_counter.add_request(
                consumer=consumer_identifier,
                method=request.method,
                path=path_template,
                status_code=response_status,
                response_time=response_time,
                request_size=request.headers.get("Content-Length"),
                response_size=response_size or response_headers.get("Content-Length"),
            )
            if response_status == 422 and response_body and response_headers.get("Content-Type") == "application/json":
                with contextlib.suppress(json.JSONDecodeError):
                    body = json.loads(response_body)
                    if isinstance(body, dict) and "detail" in body and isinstance(body["detail"], list):
                        # Log FastAPI / Pydantic validation errors
                        self.client.validation_error_counter.add_validation_errors(
                            consumer=consumer_identifier,
                            method=request.method,
                            path=path_template,
                            detail=body["detail"],
                        )
            if response_status == 500 and exception is not None:
                self.client.server_error_counter.add_server_error(
                    consumer=consumer_identifier,
                    method=request.method,
                    path=path_template,
                    exception=exception,
                )

    @staticmethod
    def get_path_template(request: Request) -> Tuple[str, bool]:
        for route in request.app.routes:
            match, _ = route.matches(request.scope)
            if match == Match.FULL:
                return route.path, True
        return request.url.path, False

    def get_consumer(self, request: Request) -> Optional[ApitallyConsumer]:
        if hasattr(request.state, "apitally_consumer") and request.state.apitally_consumer:
            return ApitallyConsumer.from_string_or_object(request.state.apitally_consumer)
        if hasattr(request.state, "consumer_identifier") and request.state.consumer_identifier:
            # Keeping this for legacy support
            warn(
                "Providing a consumer identifier via `request.state.consumer_identifier` is deprecated, "
                "use `request.state.apitally_consumer` instead.",
                DeprecationWarning,
            )
            return ApitallyConsumer.from_string_or_object(request.state.consumer_identifier)
        if self.identify_consumer_callback is not None:
            consumer = self.identify_consumer_callback(request)
            return ApitallyConsumer.from_string_or_object(consumer)
        return None


def _get_startup_data(
    app: ASGIApp, app_version: Optional[str] = None, openapi_url: Optional[str] = None
) -> Dict[str, Any]:
    data: Dict[str, Any] = {}
    if openapi_url and (openapi := _get_openapi(app, openapi_url)):
        data["openapi"] = openapi
    if endpoints := _get_endpoint_info(app):
        data["paths"] = [{"path": endpoint.path, "method": endpoint.http_method} for endpoint in endpoints]
    data["versions"] = get_versions("fastapi", "starlette", app_version=app_version)
    data["client"] = "python:starlette"
    return data


def _get_openapi(app: ASGIApp, openapi_url: str) -> Optional[str]:
    try:
        client = TestClient(app, raise_server_exceptions=False)
        response = client.get(openapi_url)
        response.raise_for_status()
        return response.text
    except HTTPStatusError:
        return None


def _get_endpoint_info(app: ASGIApp) -> List[EndpointInfo]:
    routes = _get_routes(app)
    schemas = SchemaGenerator({})
    return schemas.get_endpoints(routes)


def _get_routes(app: Union[ASGIApp, Router]) -> List[BaseRoute]:
    if isinstance(app, Router):
        return app.routes
    elif hasattr(app, "app"):
        return _get_routes(app.app)
    return []  # pragma: no cover


def _register_shutdown_handler(app: Union[ASGIApp, Router], shutdown_handler: Callable[[], Any]) -> None:
    if isinstance(app, Router):
        app.add_event_handler("shutdown", shutdown_handler)
    elif hasattr(app, "app"):
        _register_shutdown_handler(app.app, shutdown_handler)



import contextlib
import json
import time
from typing import Callable, Dict, List, Optional, Union
from warnings import warn

from litestar.app import DEFAULT_OPENAPI_CONFIG, Litestar
from litestar.config.app import AppConfig
from litestar.connection import Request
from litestar.datastructures import Headers
from litestar.enums import ScopeType
from litestar.handlers import HTTPRouteHandler
from litestar.plugins import InitPluginProtocol
from litestar.types import ASGIApp, Message, Receive, Scope, Send

from apitally.client.asyncio import ApitallyClient
from apitally.client.base import Consumer as ApitallyConsumer
from apitally.common import get_versions


__all__ = ["ApitallyPlugin", "ApitallyConsumer"]


class ApitallyPlugin(InitPluginProtocol):
    def __init__(
        self,
        client_id: str,
        env: str = "dev",
        app_version: Optional[str] = None,
        filter_openapi_paths: bool = True,
        identify_consumer_callback: Optional[Callable[[Request], Union[str, ApitallyConsumer, None]]] = None,
    ) -> None:
        self.client = ApitallyClient(client_id=client_id, env=env)
        self.app_version = app_version
        self.filter_openapi_paths = filter_openapi_paths
        self.identify_consumer_callback = identify_consumer_callback
        self.openapi_path = "/schema"

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.on_startup.append(self.on_startup)
        app_config.on_shutdown.append(self.client.handle_shutdown)
        app_config.middleware.append(self.middleware_factory)
        app_config.after_exception.append(self.after_exception)
        return app_config

    def on_startup(self, app: Litestar) -> None:
        openapi_config = app.openapi_config or DEFAULT_OPENAPI_CONFIG
        if openapi_config.openapi_controller is not None:
            self.openapi_path = openapi_config.openapi_controller.path
        elif hasattr(openapi_config, "openapi_router") and openapi_config.openapi_router is not None:
            self.openapi_path = openapi_config.openapi_router.path
        elif openapi_config.path is not None:
            self.openapi_path = openapi_config.path

        data = {
            "openapi": _get_openapi(app),
            "paths": [route for route in _get_routes(app) if not self.filter_path(route["path"])],
            "versions": get_versions("litestar", app_version=self.app_version),
            "client": "python:litestar",
        }
        self.client.set_startup_data(data)
        self.client.start_sync_loop()

    def after_exception(self, exception: Exception, scope: Scope) -> None:
        scope["state"]["exception"] = exception

    def middleware_factory(self, app: ASGIApp) -> ASGIApp:
        async def middleware(scope: Scope, receive: Receive, send: Send) -> None:
            if scope["type"] == "http" and scope["method"] != "OPTIONS":
                request = Request(scope)
                response_status = 0
                response_time = 0.0
                response_headers = Headers()
                response_body = b""
                response_size = 0
                response_chunked = False
                start_time = time.perf_counter()

                async def send_wrapper(message: Message) -> None:
                    nonlocal \
                        response_time, \
                        response_status, \
                        response_headers, \
                        response_body, \
                        response_size, \
                        response_chunked
                    if message["type"] == "http.response.start":
                        response_time = time.perf_counter() - start_time
                        response_status = message["status"]
                        response_headers = Headers(message["headers"])
                        response_chunked = (
                            response_headers.get("Transfer-Encoding") == "chunked"
                            or "Content-Length" not in response_headers
                        )
                    elif message["type"] == "http.response.body":
                        if response_chunked:
                            response_size += len(message.get("body", b""))
                        if response_status == 400:
                            response_body += message.get("body", b"")
                    await send(message)

                await app(scope, receive, send_wrapper)
                self.add_request(
                    request=request,
                    response_status=response_status,
                    response_time=response_time,
                    response_headers=response_headers,
                    response_body=response_body,
                    response_size=response_size,
                )
            else:
                await app(scope, receive, send)  # pragma: no cover

        return middleware

    def add_request(
        self,
        request: Request,
        response_status: int,
        response_time: float,
        response_headers: Headers,
        response_body: bytes,
        response_size: int = 0,
    ) -> None:
        if response_status < 100 or not request.route_handler.paths:
            return  # pragma: no cover
        path = self.get_path(request)
        if path is None or self.filter_path(path):
            return
        consumer = self.get_consumer(request)
        consumer_identifier = consumer.identifier if consumer else None
        self.client.consumer_registry.add_or_update_consumer(consumer)
        self.client.request_counter.add_request(
            consumer=consumer_identifier,
            method=request.method,
            path=path,
            status_code=response_status,
            response_time=response_time,
            request_size=request.headers.get("Content-Length"),
            response_size=response_size or response_headers.get("Content-Length"),
        )
        if response_status == 400 and response_body and len(response_body) < 4096:
            with contextlib.suppress(json.JSONDecodeError):
                parsed_body = json.loads(response_body)
                if (
                    isinstance(parsed_body, dict)
                    and "detail" in parsed_body
                    and isinstance(parsed_body["detail"], str)
                    and "validation" in parsed_body["detail"].lower()
                    and "extra" in parsed_body
                    and isinstance(parsed_body["extra"], list)
                ):
                    self.client.validation_error_counter.add_validation_errors(
                        consumer=consumer_identifier,
                        method=request.method,
                        path=path,
                        detail=[
                            {
                                "loc": [error.get("source", "body")] + error["key"].split("."),
                                "msg": error["message"],
                                "type": "",
                            }
                            for error in parsed_body["extra"]
                            if "key" in error and "message" in error
                        ],
                    )
        if response_status == 500 and "exception" in request.state:
            self.client.server_error_counter.add_server_error(
                consumer=consumer_identifier,
                method=request.method,
                path=path,
                exception=request.state["exception"],
            )

    def get_path(self, request: Request) -> Optional[str]:
        path: List[str] = []
        for layer in request.route_handler.ownership_layers:
            if isinstance(layer, HTTPRouteHandler):
                if len(layer.paths) == 0:
                    return None  # pragma: no cover
                path.append(list(layer.paths)[0].lstrip("/"))
            else:
                path.append(layer.path.lstrip("/"))
        return "/" + "/".join(filter(None, path))

    def filter_path(self, path: str) -> bool:
        if self.filter_openapi_paths and self.openapi_path:
            return path == self.openapi_path or path.startswith(self.openapi_path + "/")
        return False  # pragma: no cover

    def get_consumer(self, request: Request) -> Optional[ApitallyConsumer]:
        if hasattr(request.state, "apitally_consumer") and request.state.apitally_consumer:
            return ApitallyConsumer.from_string_or_object(request.state.apitally_consumer)
        if hasattr(request.state, "consumer_identifier") and request.state.consumer_identifier:
            # Keeping this for legacy support
            warn(
                "Providing a consumer identifier via `request.state.consumer_identifier` is deprecated, "
                "use `request.state.apitally_consumer` instead.",
                DeprecationWarning,
            )
            return ApitallyConsumer.from_string_or_object(request.state.consumer_identifier)
        if self.identify_consumer_callback is not None:
            consumer = self.identify_consumer_callback(request)
            return ApitallyConsumer.from_string_or_object(consumer)
        return None


def _get_openapi(app: Litestar) -> str:
    schema = app.openapi_schema.to_schema()
    return json.dumps(schema)


def _get_routes(app: Litestar) -> List[Dict[str, str]]:
    return [
        {"method": method, "path": route.path}
        for route in app.routes
        for method in route.methods
        if route.scope_type == ScopeType.HTTP and method != "OPTIONS"
    ]



from __future__ import annotations

import asyncio
import logging
import random
import time
from functools import partial
from typing import Any, Dict, Optional, Tuple

import backoff
import httpx

from apitally.client.base import MAX_QUEUE_TIME, REQUEST_TIMEOUT, ApitallyClientBase
from apitally.client.logging import get_logger


logger = get_logger(__name__)
retry = partial(
    backoff.on_exception,
    backoff.expo,
    httpx.HTTPError,
    max_tries=3,
    logger=logger,
    giveup_log_level=logging.WARNING,
)


class ApitallyClient(ApitallyClientBase):
    def __init__(self, client_id: str, env: str) -> None:
        super().__init__(client_id=client_id, env=env)
        self._stop_sync_loop = False
        self._sync_loop_task: Optional[asyncio.Task] = None
        self._sync_data_queue: asyncio.Queue[Tuple[float, Dict[str, Any]]] = asyncio.Queue()
        self._set_startup_data_task: Optional[asyncio.Task] = None

    def get_http_client(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(base_url=self.hub_url, timeout=REQUEST_TIMEOUT)

    def start_sync_loop(self) -> None:
        self._stop_sync_loop = False
        self._sync_loop_task = asyncio.create_task(self._run_sync_loop())

    async def _run_sync_loop(self) -> None:
        first_iteration = True
        while not self._stop_sync_loop:
            try:
                time_start = time.perf_counter()
                async with self.get_http_client() as client:
                    tasks = [self.send_sync_data(client)]
                    if not self._startup_data_sent and not first_iteration:
                        tasks.append(self.send_startup_data(client))
                    await asyncio.gather(*tasks)
                time_elapsed = time.perf_counter() - time_start
                await asyncio.sleep(self.sync_interval - time_elapsed)
            except Exception:  # pragma: no cover
                logger.exception("An error occurred during sync with Apitally hub")
            first_iteration = False

    def stop_sync_loop(self) -> None:
        self._stop_sync_loop = True

    async def handle_shutdown(self) -> None:
        if self._sync_loop_task is not None:
            self._sync_loop_task.cancel()
        # Send any remaining data before exiting
        async with self.get_http_client() as client:
            await self.send_sync_data(client)

    def set_startup_data(self, data: Dict[str, Any]) -> None:
        self._startup_data_sent = False
        self._startup_data = self.add_uuids_to_data(data)
        self._set_startup_data_task = asyncio.create_task(self._set_startup_data())

    async def _set_startup_data(self) -> None:
        async with self.get_http_client() as client:
            await self.send_startup_data(client)

    async def send_startup_data(self, client: httpx.AsyncClient) -> None:
        if self._startup_data is not None:
            await self._send_startup_data(client, self._startup_data)

    async def send_sync_data(self, client: httpx.AsyncClient) -> None:
        data = self.get_sync_data()
        self._sync_data_queue.put_nowait((time.time(), data))

        i = 0
        while not self._sync_data_queue.empty():
            timestamp, data = self._sync_data_queue.get_nowait()
            try:
                if (time_offset := time.time() - timestamp) <= MAX_QUEUE_TIME:
                    if i > 0:
                        await asyncio.sleep(random.uniform(0.1, 0.3))
                    data["time_offset"] = time_offset
                    await self._send_sync_data(client, data)
                    i += 1
            except httpx.HTTPError:
                self._sync_data_queue.put_nowait((timestamp, data))
                break
            finally:
                self._sync_data_queue.task_done()

    @retry(raise_on_giveup=False)
    async def _send_startup_data(self, client: httpx.AsyncClient, data: Dict[str, Any]) -> None:
        logger.debug("Sending startup data to Apitally hub")
        response = await client.post(url="/startup", json=data, timeout=REQUEST_TIMEOUT)
        self._handle_hub_response(response)
        self._startup_data_sent = True
        self._startup_data = None

    @retry()
    async def _send_sync_data(self, client: httpx.AsyncClient, data: Dict[str, Any]) -> None:
        logger.debug("Synchronizing data with Apitally hub")
        response = await client.post(url="/sync", json=data)
        self._handle_hub_response(response)

    def _handle_hub_response(self, response: httpx.Response) -> None:
        if response.status_code == 404:
            self.stop_sync_loop()
            logger.error("Invalid Apitally client ID: %s", self.client_id)
        elif response.status_code == 422:
            logger.error("Received validation error from Apitally hub: %s", response.json())
        else:
            response.raise_for_status()



from __future__ import annotations

import asyncio
import contextlib
import os
import re
import threading
import time
import traceback
from abc import ABC
from collections import Counter
from dataclasses import dataclass
from math import floor
from typing import Any, Dict, List, Optional, Set, Tuple, Type, TypeVar, Union, cast
from uuid import UUID, uuid4

from apitally.client.logging import get_logger


logger = get_logger(__name__)

HUB_BASE_URL = os.getenv("APITALLY_HUB_BASE_URL") or "https://hub.apitally.io"
HUB_VERSION = "v2"
REQUEST_TIMEOUT = 10
MAX_QUEUE_TIME = 3600
SYNC_INTERVAL = 60
INITIAL_SYNC_INTERVAL = 10
INITIAL_SYNC_INTERVAL_DURATION = 3600
MAX_EXCEPTION_MSG_LENGTH = 2048
MAX_EXCEPTION_TRACEBACK_LENGTH = 65536

TApitallyClient = TypeVar("TApitallyClient", bound="ApitallyClientBase")


class ApitallyClientBase(ABC):
    _instance: Optional[ApitallyClientBase] = None
    _lock = threading.Lock()

    def __new__(cls: Type[TApitallyClient], *args, **kwargs) -> TApitallyClient:
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
        return cast(TApitallyClient, cls._instance)

    def __init__(self, client_id: str, env: str) -> None:
        if hasattr(self, "client_id"):
            raise RuntimeError("Apitally client is already initialized")  # pragma: no cover
        try:
            UUID(client_id)
        except ValueError:
            raise ValueError(f"invalid client_id '{client_id}' (expecting hexadecimal UUID format)")
        if re.match(r"^[\w-]{1,32}$", env) is None:
            raise ValueError(f"invalid env '{env}' (expecting 1-32 alphanumeric lowercase characters and hyphens only)")

        self.client_id = client_id
        self.env = env
        self.instance_uuid = str(uuid4())
        self.request_counter = RequestCounter()
        self.validation_error_counter = ValidationErrorCounter()
        self.server_error_counter = ServerErrorCounter()
        self.consumer_registry = ConsumerRegistry()

        self._startup_data: Optional[Dict[str, Any]] = None
        self._startup_data_sent = False
        self._started_at = time.time()

    @classmethod
    def get_instance(cls: Type[TApitallyClient]) -> TApitallyClient:
        if cls._instance is None:
            raise RuntimeError("Apitally client not initialized")  # pragma: no cover
        return cast(TApitallyClient, cls._instance)

    @property
    def sync_interval(self) -> float:
        return (
            SYNC_INTERVAL if time.time() - self._started_at > INITIAL_SYNC_INTERVAL_DURATION else INITIAL_SYNC_INTERVAL
        )

    @property
    def hub_url(self) -> str:
        return f"{HUB_BASE_URL}/{HUB_VERSION}/{self.client_id}/{self.env}"

    def add_uuids_to_data(self, data: Dict[str, Any]) -> Dict[str, Any]:
        data_with_uuids = {
            "instance_uuid": self.instance_uuid,
            "message_uuid": str(uuid4()),
        }
        data_with_uuids.update(data)
        return data_with_uuids

    def get_sync_data(self) -> Dict[str, Any]:
        data = {
            "requests": self.request_counter.get_and_reset_requests(),
            "validation_errors": self.validation_error_counter.get_and_reset_validation_errors(),
            "server_errors": self.server_error_counter.get_and_reset_server_errors(),
            "consumers": self.consumer_registry.get_and_reset_updated_consumers(),
        }
        return self.add_uuids_to_data(data)


@dataclass(frozen=True)
class RequestInfo:
    consumer: Optional[str]
    method: str
    path: str
    status_code: int


class RequestCounter:
    def __init__(self) -> None:
        self.request_counts: Counter[RequestInfo] = Counter()
        self.request_size_sums: Counter[RequestInfo] = Counter()
        self.response_size_sums: Counter[RequestInfo] = Counter()
        self.response_times: Dict[RequestInfo, Counter[int]] = {}
        self.request_sizes: Dict[RequestInfo, Counter[int]] = {}
        self.response_sizes: Dict[RequestInfo, Counter[int]] = {}
        self._lock = threading.Lock()

    def add_request(
        self,
        consumer: Optional[str],
        method: str,
        path: str,
        status_code: int,
        response_time: float,
        request_size: str | int | None = None,
        response_size: str | int | None = None,
    ) -> None:
        request_info = RequestInfo(
            consumer=consumer,
            method=method.upper(),
            path=path,
            status_code=status_code,
        )
        response_time_ms_bin = int(floor(response_time / 0.01) * 10)  # In ms, rounded down to nearest 10ms
        with self._lock:
            self.request_counts[request_info] += 1
            self.response_times.setdefault(request_info, Counter())[response_time_ms_bin] += 1
            if request_size is not None:
                with contextlib.suppress(ValueError):
                    request_size = int(request_size)
                    request_size_kb_bin = request_size // 1000  # In KB, rounded down to nearest 1KB
                    self.request_size_sums[request_info] += request_size
                    self.request_sizes.setdefault(request_info, Counter())[request_size_kb_bin] += 1
            if response_size is not None:
                with contextlib.suppress(ValueError):
                    response_size = int(response_size)
                    response_size_kb_bin = response_size // 1000  # In KB, rounded down to nearest 1KB
                    self.response_size_sums[request_info] += response_size
                    self.response_sizes.setdefault(request_info, Counter())[response_size_kb_bin] += 1

    def get_and_reset_requests(self) -> List[Dict[str, Any]]:
        data: List[Dict[str, Any]] = []
        with self._lock:
            for request_info, count in self.request_counts.items():
                data.append(
                    {
                        "consumer": request_info.consumer,
                        "method": request_info.method,
                        "path": request_info.path,
                        "status_code": request_info.status_code,
                        "request_count": count,
                        "request_size_sum": self.request_size_sums.get(request_info, 0),
                        "response_size_sum": self.response_size_sums.get(request_info, 0),
                        "response_times": self.response_times.get(request_info) or Counter(),
                        "request_sizes": self.request_sizes.get(request_info) or Counter(),
                        "response_sizes": self.response_sizes.get(request_info) or Counter(),
                    }
                )
            self.request_counts.clear()
            self.request_size_sums.clear()
            self.response_size_sums.clear()
            self.response_times.clear()
            self.request_sizes.clear()
            self.response_sizes.clear()
        return data


@dataclass(frozen=True)
class ValidationError:
    consumer: Optional[str]
    method: str
    path: str
    loc: Tuple[str, ...]
    msg: str
    type: str


class ValidationErrorCounter:
    def __init__(self) -> None:
        self.error_counts: Counter[ValidationError] = Counter()
        self._lock = threading.Lock()

    def add_validation_errors(
        self, consumer: Optional[str], method: str, path: str, detail: List[Dict[str, Any]]
    ) -> None:
        with self._lock:
            for error in detail:
                try:
                    validation_error = ValidationError(
                        consumer=consumer,
                        method=method.upper(),
                        path=path,
                        loc=tuple(str(loc) for loc in error["loc"]),
                        msg=error["msg"],
                        type=error["type"],
                    )
                    self.error_counts[validation_error] += 1
                except (KeyError, TypeError):  # pragma: no cover
                    pass

    def get_and_reset_validation_errors(self) -> List[Dict[str, Any]]:
        data: List[Dict[str, Any]] = []
        with self._lock:
            for validation_error, count in self.error_counts.items():
                data.append(
                    {
                        "consumer": validation_error.consumer,
                        "method": validation_error.method,
                        "path": validation_error.path,
                        "loc": validation_error.loc,
                        "msg": validation_error.msg,
                        "type": validation_error.type,
                        "error_count": count,
                    }
                )
            self.error_counts.clear()
        return data


@dataclass(frozen=True)
class ServerError:
    consumer: Optional[str]
    method: str
    path: str
    type: str
    msg: str
    traceback: str


class ServerErrorCounter:
    def __init__(self) -> None:
        self.error_counts: Counter[ServerError] = Counter()
        self.sentry_event_ids: Dict[ServerError, str] = {}
        self._lock = threading.Lock()
        self._tasks: Set[asyncio.Task] = set()

    def add_server_error(self, consumer: Optional[str], method: str, path: str, exception: BaseException) -> None:
        if not isinstance(exception, BaseException):
            return  # pragma: no cover
        exception_type = type(exception)
        with self._lock:
            server_error = ServerError(
                consumer=consumer,
                method=method.upper(),
                path=path,
                type=f"{exception_type.__module__}.{exception_type.__qualname__}",
                msg=self._get_truncated_exception_msg(exception),
                traceback=self._get_truncated_exception_traceback(exception),
            )
            self.error_counts[server_error] += 1
            self.capture_sentry_event_id(server_error)

    def capture_sentry_event_id(self, server_error: ServerError) -> None:
        try:
            from sentry_sdk.hub import Hub
            from sentry_sdk.scope import Scope
        except ImportError:
            return  # pragma: no cover
        if not hasattr(Scope, "get_isolation_scope") or not hasattr(Scope, "_last_event_id"):
            # sentry-sdk < 2.2.0 is not supported
            return  # pragma: no cover
        if Hub.current.client is None:
            return  # sentry-sdk not initialized

        scope = Scope.get_isolation_scope()
        if event_id := scope._last_event_id:
            self.sentry_event_ids[server_error] = event_id
            return

        async def _wait_for_sentry_event_id(scope: Scope) -> None:
            i = 0
            while not (event_id := scope._last_event_id) and i < 100:
                i += 1
                await asyncio.sleep(0.001)
            if event_id:
                self.sentry_event_ids[server_error] = event_id

        with contextlib.suppress(RuntimeError):  # ignore no running loop
            loop = asyncio.get_running_loop()
            task = loop.create_task(_wait_for_sentry_event_id(scope))
            self._tasks.add(task)
            task.add_done_callback(self._tasks.discard)

    def get_and_reset_server_errors(self) -> List[Dict[str, Any]]:
        data: List[Dict[str, Any]] = []
        with self._lock:
            for server_error, count in self.error_counts.items():
                data.append(
                    {
                        "consumer": server_error.consumer,
                        "method": server_error.method,
                        "path": server_error.path,
                        "type": server_error.type,
                        "msg": server_error.msg,
                        "traceback": server_error.traceback,
                        "sentry_event_id": self.sentry_event_ids.get(server_error),
                        "error_count": count,
                    }
                )
            self.error_counts.clear()
            self.sentry_event_ids.clear()
        return data

    @staticmethod
    def _get_truncated_exception_msg(exception: BaseException) -> str:
        msg = str(exception).strip()
        if len(msg) <= MAX_EXCEPTION_MSG_LENGTH:
            return msg
        suffix = "... (truncated)"
        cutoff = MAX_EXCEPTION_MSG_LENGTH - len(suffix)
        return msg[:cutoff] + suffix

    @staticmethod
    def _get_truncated_exception_traceback(exception: BaseException) -> str:
        prefix = "... (truncated) ...\n"
        cutoff = MAX_EXCEPTION_TRACEBACK_LENGTH - len(prefix)
        lines = []
        length = 0
        for line in traceback.format_exception(exception)[::-1]:
            if length + len(line) > cutoff:
                lines.append(prefix)
                break
            lines.append(line)
            length += len(line)
        return "".join(lines[::-1]).strip()


class Consumer:
    def __init__(self, identifier: str, name: Optional[str] = None, group: Optional[str] = None) -> None:
        self.identifier = str(identifier).strip()[:128]
        self.name = str(name).strip()[:64] if name else None
        self.group = str(group).strip()[:64] if group else None

    @classmethod
    def from_string_or_object(cls, consumer: Optional[Union[str, Consumer]]) -> Optional[Consumer]:
        if not consumer:
            return None
        if isinstance(consumer, Consumer):
            return consumer
        consumer = str(consumer).strip()
        if not consumer:
            return None
        return cls(identifier=consumer)

    def update(self, name: str | None = None, group: str | None = None) -> bool:
        name = str(name).strip()[:64] if name else None
        group = str(group).strip()[:64] if group else None
        updated = False
        if name and name != self.name:
            self.name = name
            updated = True
        if group and group != self.group:
            self.group = group
            updated = True
        return updated


class ConsumerRegistry:
    def __init__(self) -> None:
        self.consumers: Dict[str, Consumer] = {}
        self.updated: Set[str] = set()
        self._lock = threading.Lock()

    def add_or_update_consumer(self, consumer: Optional[Consumer]) -> None:
        if not consumer or (not consumer.name and not consumer.group):
            return  # Only register consumers with name or group set
        with self._lock:
            if consumer.identifier not in self.consumers:
                self.consumers[consumer.identifier] = consumer
                self.updated.add(consumer.identifier)
            elif self.consumers[consumer.identifier].update(name=consumer.name, group=consumer.group):
                self.updated.add(consumer.identifier)

    def get_and_reset_updated_consumers(self) -> List[Dict[str, Any]]:
        data: List[Dict[str, Any]] = []
        with self._lock:
            for identifier in self.updated:
                if consumer := self.consumers.get(identifier):
                    data.append(
                        {
                            "identifier": consumer.identifier,
                            "name": str(consumer.name)[:64] if consumer.name else None,
                            "group": str(consumer.group)[:64] if consumer.group else None,
                        }
                    )
            self.updated.clear()
        return data



from __future__ import annotations

import time
from threading import Timer
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Tuple
from warnings import warn

from flask import Flask, g
from flask.wrappers import Response
from werkzeug.datastructures import Headers
from werkzeug.exceptions import NotFound
from werkzeug.test import Client

from apitally.client.base import Consumer as ApitallyConsumer
from apitally.client.threading import ApitallyClient
from apitally.common import get_versions


if TYPE_CHECKING:
    from _typeshed.wsgi import StartResponse, WSGIApplication, WSGIEnvironment
    from werkzeug.routing.map import Map


__all__ = ["ApitallyMiddleware", "ApitallyConsumer"]


class ApitallyMiddleware:
    def __init__(
        self,
        app: Flask,
        client_id: str,
        env: str = "dev",
        app_version: Optional[str] = None,
        openapi_url: Optional[str] = None,
        filter_unhandled_paths: bool = True,
    ) -> None:
        self.app = app
        self.wsgi_app = app.wsgi_app
        self.filter_unhandled_paths = filter_unhandled_paths
        self.patch_handle_exception()
        self.client = ApitallyClient(client_id=client_id, env=env)
        self.client.start_sync_loop()
        self.delayed_set_startup_data(app_version, openapi_url)

    def delayed_set_startup_data(self, app_version: Optional[str] = None, openapi_url: Optional[str] = None) -> None:
        # Short delay to allow app routes to be registered first
        timer = Timer(
            1.0,
            self._delayed_set_startup_data,
            kwargs={"app_version": app_version, "openapi_url": openapi_url},
        )
        timer.start()

    def _delayed_set_startup_data(self, app_version: Optional[str] = None, openapi_url: Optional[str] = None) -> None:
        data = _get_startup_data(self.app, app_version, openapi_url)
        self.client.set_startup_data(data)

    def __call__(self, environ: WSGIEnvironment, start_response: StartResponse) -> Iterable[bytes]:
        status_code = 200
        response_headers = Headers([])

        def catching_start_response(status: str, headers: List[Tuple[str, str]], exc_info=None):
            nonlocal status_code, response_headers
            status_code = int(status.split(" ")[0])
            response_headers = Headers(headers)
            return start_response(status, headers, exc_info)

        start_time = time.perf_counter()
        with self.app.app_context():
            response = self.wsgi_app(environ, catching_start_response)
            self.add_request(
                environ=environ,
                status_code=status_code,
                response_time=time.perf_counter() - start_time,
                response_headers=response_headers,
            )
        return response

    def patch_handle_exception(self) -> None:
        original_handle_exception = self.app.handle_exception

        def handle_exception(e: Exception) -> Response:
            g.unhandled_exception = e
            return original_handle_exception(e)

        self.app.handle_exception = handle_exception  # type: ignore[method-assign]

    def add_request(
        self,
        environ: WSGIEnvironment,
        status_code: int,
        response_time: float,
        response_headers: Headers,
    ) -> None:
        rule, is_handled_path = self.get_rule(environ)
        if (is_handled_path or not self.filter_unhandled_paths) and environ["REQUEST_METHOD"] != "OPTIONS":
            consumer = self.get_consumer()
            consumer_identifier = consumer.identifier if consumer else None
            self.client.consumer_registry.add_or_update_consumer(consumer)
            self.client.request_counter.add_request(
                consumer=consumer_identifier,
                method=environ["REQUEST_METHOD"],
                path=rule,
                status_code=status_code,
                response_time=response_time,
                request_size=environ.get("CONTENT_LENGTH"),
                response_size=response_headers.get("Content-Length", type=int),
            )
            if status_code == 500 and "unhandled_exception" in g:
                self.client.server_error_counter.add_server_error(
                    consumer=consumer_identifier,
                    method=environ["REQUEST_METHOD"],
                    path=rule,
                    exception=g.unhandled_exception,
                )

    def get_rule(self, environ: WSGIEnvironment) -> Tuple[str, bool]:
        url_adapter = self.app.url_map.bind_to_environ(environ)
        try:
            endpoint, _ = url_adapter.match()
            rule = self.app.url_map._rules_by_endpoint[endpoint][0]
            return rule.rule, True
        except NotFound:
            return environ["PATH_INFO"], False

    def get_consumer(self) -> Optional[ApitallyConsumer]:
        if "apitally_consumer" in g and g.apitally_consumer:
            return ApitallyConsumer.from_string_or_object(g.apitally_consumer)
        if "consumer_identifier" in g and g.consumer_identifier:
            # Keeping this for legacy support
            warn(
                "Providing a consumer identifier via `g.consumer_identifier` is deprecated, "
                "use `g.apitally_consumer` instead.",
                DeprecationWarning,
            )
            return ApitallyConsumer.from_string_or_object(g.consumer_identifier)
        return None


def _get_startup_data(
    app: Flask, app_version: Optional[str] = None, openapi_url: Optional[str] = None
) -> Dict[str, Any]:
    data: Dict[str, Any] = {}
    if openapi_url and (openapi := _get_openapi(app, openapi_url)):
        data["openapi"] = openapi
    if paths := _get_paths(app.url_map):
        data["paths"] = paths
    data["versions"] = get_versions("flask", app_version=app_version)
    data["client"] = "python:flask"
    return data


def _get_paths(url_map: Map) -> List[Dict[str, str]]:
    return [
        {"path": rule.rule, "method": method}
        for rule in url_map.iter_rules()
        if rule.methods is not None and rule.rule != "/static/<path:filename>"
        for method in rule.methods
        if method not in ["HEAD", "OPTIONS"]
    ]


def _get_openapi(app: WSGIApplication, openapi_url: str) -> Optional[str]:
    client = Client(app)
    response = client.get(openapi_url)
    if response.status_code != 200:
        return None
    return response.get_data(as_text=True)



from __future__ import annotations

import logging
import queue
import random
import time
from functools import partial
from threading import Event, Thread
from typing import Any, Callable, Dict, Optional, Tuple

import backoff
import requests

from apitally.client.base import MAX_QUEUE_TIME, REQUEST_TIMEOUT, ApitallyClientBase
from apitally.client.logging import get_logger


logger = get_logger(__name__)
retry = partial(
    backoff.on_exception,
    backoff.expo,
    requests.RequestException,
    max_tries=3,
    logger=logger,
    giveup_log_level=logging.WARNING,
)


# Function to register an on-exit callback for both Python and IPython runtimes
try:

    def register_exit(func: Callable[..., Any], *args, **kwargs) -> Callable[..., Any]:  # pragma: no cover
        def callback():
            func()
            ipython.events.unregister("post_execute", callback)

        ipython.events.register("post_execute", callback)
        return func

    ipython = get_ipython()  # type: ignore
except NameError:
    from atexit import register as register_exit  # type: ignore[assignment]


class ApitallyClient(ApitallyClientBase):
    def __init__(self, client_id: str, env: str) -> None:
        super().__init__(client_id=client_id, env=env)
        self._thread: Optional[Thread] = None
        self._stop_sync_loop = Event()
        self._sync_data_queue: queue.Queue[Tuple[float, Dict[str, Any]]] = queue.Queue()

    def start_sync_loop(self) -> None:
        self._stop_sync_loop.clear()
        if self._thread is None or not self._thread.is_alive():
            self._thread = Thread(target=self._run_sync_loop, daemon=True)
            self._thread.start()
            register_exit(self.stop_sync_loop)

    def _run_sync_loop(self) -> None:
        try:
            last_sync_time = 0.0
            while not self._stop_sync_loop.is_set():
                try:
                    now = time.time()
                    if (now - last_sync_time) >= self.sync_interval:
                        with requests.Session() as session:
                            if not self._startup_data_sent and last_sync_time > 0:  # not on first sync
                                self.send_startup_data(session)
                            self.send_sync_data(session)
                        last_sync_time = now
                    time.sleep(1)
                except Exception:  # pragma: no cover
                    logger.exception("An error occurred during sync with Apitally hub")
        finally:
            # Send any remaining data before exiting
            with requests.Session() as session:
                self.send_sync_data(session)

    def stop_sync_loop(self) -> None:
        self._stop_sync_loop.set()
        if self._thread is not None:
            self._thread.join()
            self._thread = None

    def set_startup_data(self, data: Dict[str, Any]) -> None:
        self._startup_data_sent = False
        self._startup_data = self.add_uuids_to_data(data)
        with requests.Session() as session:
            self.send_startup_data(session)

    def send_startup_data(self, session: requests.Session) -> None:
        if self._startup_data is not None:
            self._send_startup_data(session, self._startup_data)

    def send_sync_data(self, session: requests.Session) -> None:
        data = self.get_sync_data()
        self._sync_data_queue.put_nowait((time.time(), data))

        i = 0
        while not self._sync_data_queue.empty():
            timestamp, data = self._sync_data_queue.get_nowait()
            try:
                if (time_offset := time.time() - timestamp) <= MAX_QUEUE_TIME:
                    if i > 0:
                        time.sleep(random.uniform(0.1, 0.3))
                    data["time_offset"] = time_offset
                    self._send_sync_data(session, data)
                    i += 1
            except requests.RequestException:
                self._sync_data_queue.put_nowait((timestamp, data))
                break
            finally:
                self._sync_data_queue.task_done()

    @retry(raise_on_giveup=False)
    def _send_startup_data(self, session: requests.Session, data: Dict[str, Any]) -> None:
        logger.debug("Sending startup data to Apitally hub")
        response = session.post(url=f"{self.hub_url}/startup", json=data, timeout=REQUEST_TIMEOUT)
        self._handle_hub_response(response)
        self._startup_data_sent = True
        self._startup_data = None

    @retry()
    def _send_sync_data(self, session: requests.Session, data: Dict[str, Any]) -> None:
        logger.debug("Synchronizing data with Apitally hub")
        response = session.post(url=f"{self.hub_url}/sync", json=data, timeout=REQUEST_TIMEOUT)
        self._handle_hub_response(response)

    def _handle_hub_response(self, response: requests.Response) -> None:
        if response.status_code == 404:
            self.stop_sync_loop()
            logger.error("Invalid Apitally client ID: %s", self.client_id)
        elif response.status_code == 422:
            logger.error("Received validation error from Apitally hub: %s", response.json())
        else:
            response.raise_for_status()
